var should = require('should'),
    sys = require('sys'),
    db = require('../lib/fakedb').init('search.fdb');

var data = [
  { key: 'key-0', username: 'jaydee', password: 'secret7' },
  { key: 'key-2', username: 'addie', password: 'secret3' },
  { key: 'key-3', username: 'oliver', password: 'secret1' },
  { key: 'key-5', username: 'andee', password: 'secret5' },
  { key: 'key-1', username: 'sheryl', password: 'secret5' },
  { key: 'key-6', username: 'becky', password: 'secret7' },
  { key: 'key-4', username: 'aleksees', password: 'secret5' },
  { key: 'key-7', username: 'ernesto', password: 'secret5' }
];

var TESTS = {

  'Testing db.set()': function () {
    var i = 0;
    data.forEach(function (user) {
      db.set(user.key, user);
      ++i;
    });
    
    db.index('username');
    db.index('password');

    db.count().should.equal(data.length);
  },

  'Testing db.get()': function () {
    data.forEach(function (obj) {
      db.get(obj.key, function (key, value) {
        should.exist(key);
        should.exist(value);
        value.should.have.property('username', obj.username);
        value.should.have.property('password', obj.password);
      });
    });
  },

  'Testing db.all() with just a callback': function () {
    db.all(function (matches) {
      should.exist(matches);
      matches.should.have.property('length', db.count());
    });
  },

  'Testing db.all() with options': function () {
    db.all({ offset: 0, limit: 2 }, function (matches) {
      should.exist(matches);
      matches.should.have.property('length', 2);
    });
  },

  'Testing db.first()': function () {
    db.first(function (key, value) {
      should.exist(key);
      should.exist(value);
      value.should.have.property('username', 'jaydee');
    });
  },

  'Testing db.last()': function () {
    db.last(function (key, value) {
      should.exist(key);
      should.exist(value);
      value.should.have.property('username', 'ernesto');
    });
  },

  'Testing db.first()': function () {
    db.first(function (key, value) {
      should.exist(value);
      value.should.have.property('username', 'jaydee');
    });
  },

  'Testing db.last()': function () {
    db.last(function (key, value) {
      should.exist(value);
      value.should.have.property('username', 'ernesto');
    });
  },

  'Testing db.filter()': function () {
    var checker = function (o) {
      return (o.username == 'addie');
    };
    db.filter(checker, function (o) {
      should.exist(o);
      o.should.have.property('username', 'addie');
    });
  },

  'Testing db.search()': function () {
    db.search({ username: 'addie' }, function (matches) {
      should.exist(matches);
      matches.should.have.property('length', 1);
      matches[0].should.have.property('username', 'addie');
    });
  },
  
  'Testing db.search() with compound query': function () {
    db.search({ username: 'addie', password: 'secret3' }, function (matches) {
      should.exist(matches);
      matches.should.have.property('length', 1);
      matches[0].should.have.property('username', 'addie');
    });
  },

  'Testing db.search() with compound query for non-existent record': function () {
    db.search({ username: 'noodlehaus', password: 'secretnoodle' }, function (matches) {
      should.exist(matches);
    });
  },

  'Testing db.search() on non-existent record': function () {
    db.search({ username: 'jesus' }, function (matches) {
      should.exist(matches);
      matches.should.have.property('length', 0);
    });
  },

  'Testing db.search() with a modified index': function () {
    db.set('key-x', {
      key: 'key-x',
      username: 'noodlehaus',
      password: 'secretX'
    });
    db.search({ username: 'noodlehaus' }, function (matches) {
      should.exist(matches);
      matches.should.have.property('length', 1);
      matches[0].should.have.property('username', 'noodlehaus');
    });
    db.remove('key-x');
  },

  'Testing db.search() with options': function () {
    db.search({ password: 'secret5' }, { offset: 1, limit: 2 }, function (matches) {
      should.exist(matches);
      matches.should.have.property('length', 2);
    });
  },

  'Testing db.remove()': function () {
    db.all(function (key, value) {
      db.remove(key);
      db.get(key, function (key, value) {
        should.not.exist(key);
        should.not.exist(value);
      });
    });
  }

};

function red(str) {
  return '\033[31m' + str + '\033[0m';
}

function green(str) {
  return '\033[32m' + str + '\033[0m';
}

db.on('load', function () {

  var total = 0,
      passed = 0,
      failed = 0;

  console.log('DB (' + db.path + ') loaded. Starting tests.');
  for (var spec in TESTS) {
    try {
      sys.print(spec, '... ');
      TESTS[spec].call(undefined);
      console.log(green('PASSED'));
      total++;
      passed++;
    } catch (e) {
      console.log(red('FAILED'));
      console.log(red('-- ' + e));
      failed++;
    }
  }

  console.log(
    'FINISHED\n' +
    'Total: ' + total +
    ', Passed: ' + passed +
    ', Failed: ' + failed
  );

});
